import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import patches
import lmfit
from lmfit.models import GaussianModel, VoigtModel, LinearModel, ConstantModel
from scipy.signal import find_peaks
import os
import re
from os import listdir
from os.path import isfile, join
import pickle
from pathlib import Path
from DiadFit.ne_lines import *
from scipy.stats import t
from DiadFit.importing_data_files import *
encode="ISO-8859-1"

## Cornell densimeters
def calculate_density_cornell_old(*, temp='SupCrit', Split, split_err=None):
    """ This function converts Diad Splitting into CO$_2$ density using the densimeters of DeVitre et al. (2021)
    This should only be used for the Cornell Raman, not other Ramans at present. This is an older version of the function
    calculate_density_cornell - it does not have the same error propagation capabilities, but we keep it here for backwards
    compatability.

    Parameters
    -------------
    temp: str
        'SupCrit' if measurements done at 37C
        'RoomT' if measurements done at 24C

    Split: int, float, pd.Series, np.array
        Corrected splitting in cm-1

    Split_err: int, float, pd.Series (Optional)
        Error on corrected splitting in cm-1


    Returns
    --------------
    pd.DataFrame
        Prefered Density (based on different equations being merged), and intermediate calculations




    """

    #if temp is "RoomT":
    LowD_RT=-38.34631 + 0.3732578*Split
    HighD_RT=-41.64784 + 0.4058777*Split- 0.1460339*(Split-104.653)**2

    # IF temp is 37
    LowD_SC=-38.62718 + 0.3760427*Split
    MedD_SC=-47.2609 + 0.4596005*Split+ 0.0374189*(Split-103.733)**2-0.0187173*(Split-103.733)**3
    HighD_SC=-42.52782 + 0.4144277*Split- 0.1514429*(Split-104.566)**2

    if isinstance(Split, pd.Series) or isinstance(Split, np.ndarray):

        df=pd.DataFrame(data={'Preferred D': 0,
        'in range': 'Y',
        'Corrected_Splitting': Split,
                                'Notes': 'not in range',
                                'LowD_RT': LowD_RT,
                                'HighD_RT': HighD_RT,
                                'LowD_SC': LowD_SC,
                                'MedD_SC': MedD_SC,
                                'HighD_SC': HighD_SC,
                                'Temperature': temp,
                                'Splitting': Split,

                                })

    else:
        df=pd.DataFrame(data={'Preferred D': 0,
        'in range': 'Y',
        'Corrected_Splitting': Split,
                                'Notes': 'not in range',
                                'LowD_RT': LowD_RT,
                                'HighD_RT': HighD_RT,
                                'LowD_SC': LowD_SC,
                                'MedD_SC': MedD_SC,
                                'HighD_SC': HighD_SC,
                                'Temperature': temp,
                                'Splitting': Split,

                                }, index=[0])


    roomT=df['Temperature']=="RoomT"
    SupCrit=df['Temperature']=="SupCrit"
    # If splitting is 0
    zero=df['Corrected_Splitting']==0

    # Range for SC low density
    min_lowD_SC_Split=df['Corrected_Splitting']>=102.72
    max_lowD_SC_Split=df['Corrected_Splitting']<=103.16
    # Range for SC med density
    min_MD_SC_Split=df['Corrected_Splitting']>103.16
    max_MD_SC_Split=df['Corrected_Splitting']<=104.28
    # Range for SC high density
    min_HD_SC_Split=df['Corrected_Splitting']>=104.28
    max_HD_SC_Split=df['Corrected_Splitting']<=104.95
    # Range for Room T low density
    min_lowD_RoomT_Split=df['Corrected_Splitting']>=102.734115670188
    max_lowD_RoomT_Split=df['Corrected_Splitting']<=103.350311768435
    # Range for Room T high density
    min_HD_RoomT_Split=df['Corrected_Splitting']>=104.407308904012
    max_HD_RoomT_Split=df['Corrected_Splitting']<=105.1
    # Impossible densities, room T
    Imposs_lower_end=(df['Corrected_Splitting']>103.350311768435) & (df['Splitting']<103.88)
    # Impossible densities, room T
    Imposs_upper_end=(df['Corrected_Splitting']<104.407308904012) & (df['Splitting']>103.88)
    # Too low density
    Too_Low_SC=df['Corrected_Splitting']<102.72
    Too_Low_RT=df['Corrected_Splitting']<102.734115670188

    df.loc[zero, 'Preferred D']=0
    df.loc[zero, 'Notes']=0


    # If room T, low density, set as low density
    df.loc[roomT&(min_lowD_RoomT_Split&max_lowD_RoomT_Split), 'Preferred D'] = LowD_RT
    df.loc[roomT&(min_lowD_RoomT_Split&max_lowD_RoomT_Split), 'Notes']='Room T, low density'
    # If room T, high density
    df.loc[roomT&(min_HD_RoomT_Split&max_HD_RoomT_Split), 'Preferred D'] = HighD_RT
    df.loc[roomT&(min_HD_RoomT_Split&max_HD_RoomT_Split), 'Notes']='Room T, high density'

    # If SupCrit, high density
    df.loc[ SupCrit&(min_HD_SC_Split&max_HD_SC_Split), 'Preferred D'] = HighD_SC
    df.loc[ SupCrit&(min_HD_SC_Split&max_HD_SC_Split), 'Notes']='SupCrit, high density'
    # If SupCrit, Med density
    df.loc[SupCrit&(min_MD_SC_Split&max_MD_SC_Split), 'Preferred D'] = MedD_SC
    df.loc[SupCrit&(min_MD_SC_Split&max_MD_SC_Split), 'Notes']='SupCrit, Med density'

    # If SupCrit, low density
    df.loc[ SupCrit&(min_lowD_SC_Split&max_lowD_SC_Split), 'Preferred D'] = LowD_SC
    df.loc[SupCrit&(min_lowD_SC_Split&max_lowD_SC_Split), 'Notes']='SupCrit, low density'

    # If Supcritical, and too low
    df.loc[SupCrit&(Too_Low_SC), 'Preferred D']=LowD_SC
    df.loc[SupCrit&(Too_Low_SC), 'Notes']='Below lower calibration limit'
    df.loc[SupCrit&(Too_Low_SC), 'in range']='N'


    # If RoomT, and too low
    df.loc[roomT&(Too_Low_RT), 'Preferred D']=LowD_RT
    df.loc[roomT&(Too_Low_RT), 'Notes']='Below lower calibration limit'
    df.loc[roomT&(Too_Low_RT), 'in range']='N'

    #if splitting is zero
    SplitZero=df['Corrected_Splitting']==0
    df.loc[SupCrit&(SplitZero), 'Preferred D']=np.nan
    df.loc[SupCrit&(SplitZero), 'Notes']='Splitting=0'
    df.loc[SupCrit&(SplitZero), 'in range']='N'

    df.loc[roomT&(SplitZero), 'Preferred D']=np.nan
    df.loc[roomT&(SplitZero), 'Notes']='Splitting=0'
    df.loc[roomT&(SplitZero), 'in range']='N'


    # If impossible density, lower end
    df.loc[roomT&Imposs_lower_end, 'Preferred D'] = LowD_RT
    df.loc[roomT&Imposs_lower_end, 'Notes']='Impossible Density, low density'
    df.loc[roomT&Imposs_lower_end, 'in range']='N'

    # If impossible density, lower end
    df.loc[roomT&Imposs_upper_end, 'Preferred D'] = HighD_RT
    df.loc[roomT&Imposs_upper_end, 'Notes']='Impossible Density, high density'
    df.loc[roomT&Imposs_upper_end, 'in range']='N'


    #df.loc[zero, 'in range']='Y'
    # If high densiy, and beyond the upper calibration limit
    Upper_Cal_RT=df['Corrected_Splitting']>105.1
    Upper_Cal_SC=df['Corrected_Splitting']>104.95

    df.loc[roomT&Upper_Cal_RT, 'Preferred D'] = HighD_RT
    df.loc[roomT&Upper_Cal_RT, 'Notes']='Above upper Cali Limit'
    df.loc[roomT&Upper_Cal_RT, 'in range']='N'

    df.loc[SupCrit&Upper_Cal_SC, 'Preferred D'] = HighD_SC
    df.loc[SupCrit&Upper_Cal_SC, 'Notes']='Above upper Cali Limit'
    df.loc[SupCrit&Upper_Cal_SC, 'in range']='N'

    if split_err is not None:
        df2=calculate_dens_error(temp, Split, split_err)

        df.insert(1, 'dens+1σ', df2['max_dens'])
        df.insert(1, 'dens-1σ', df2['min_dens'])
        df.insert(3, '1σ', (df2['max_dens']-df2['min_dens'])/2 )

    return df

def calculate_dens_error(temp, Split, split_err):
    """
    This function is used by the function calculate_density_cornell_old to calculate a max and min density
    corresponding to splitting values + - 1 sigma

    Parameters
    -------------
    temp: str
        'SupCrit' if measurements done at 37C
        'RoomT' if measurements done at 24C

    Split: int, float, pd.Series, np.array
        Corrected splitting in cm-1

    Split_err: int, float, pd.Series (Optional)
        Error on corrected splitting in cm-1

    Returns
    ---------
    pd.DataFrame
        Dataframe with column for max calculated density and min calculated density.
    """

    max_dens=calculate_density_cornell(temp=temp, Split=Split+split_err)
    min_dens=calculate_density_cornell(temp=temp, Split=Split-split_err)
    df=pd.DataFrame(data={
                        'max_dens': max_dens['Preferred D'],
                        'min_dens': min_dens['Preferred D']})

    return df




def propagate_error_split_neon_peakfit(*, df_fits, Ne_corr=None, Ne_err=None, pref_Ne=None):
    """ This function propagates errors in your Ne correction model and peak fits by quadrature.

    Parameters
    -----------------

    df_fits: pd.DataFrame
        Dataframe of peak fitting parameters. Must contain columns for 'Diad1_cent_err', 'Diad2_cent_err', 'Splitting'

    Choose either:

    Ne_corr: pd.DataFrame (Optional)
        Dataframe with columns for 'upper_values' and 'lower values', e.g. the upper and lower bounds of the error on the Ne correction model

    Or

    pref_Ne and Ne_err: float, int, pd.Series, np.array
        A preferred value of the Ne correction factor and the error (e.g. pref_Ne=0.998, Ne_err=0.001). Used for
        rapid peak fitting before developing Ne lines.


    Returns
    -----------------
    two pd.Series, the error on the splitting, and the combined error from the splitting and the Ne correction model.

    """
    # Get the error on Neon things
    if isinstance(Ne_corr, pd.DataFrame):
        Ne_err=(Ne_corr['upper_values']-Ne_corr['lower_values'])/2
        print(np.mean(Ne_err))
        pref_Ne=Ne_corr['preferred_values']

    elif pref_Ne is not None and Ne_err is not None:
        print('using fixed values for Ne error and Ne factor')
    else:
        raise TypeError('you either ne Ne_corr as a dataframe, or to give a value for pref_Ne and Ne_err')




    # Get the peak fit errors
    Diad1_err=df_fits['Diad1_cent_err'].fillna(0)
    Diad2_err=df_fits['Diad2_cent_err'].fillna(0)
    split_err=(Diad1_err**2 + Diad2_err**2)**0.5
    Combo_err= (((df_fits['Splitting']* (Ne_err))**2) +  (pref_Ne *split_err  )**2 )**0.5




    return Combo_err, split_err

## Error for densimeters

def calculate_Densimeter_std_err_values(*, pickle_str, corrected_split, corrected_split_err, CI_dens=0.67, CI_split=0.67, str_d='LowD') :

    """
    This function propagates uncertainty from the densimeter polynomial fit and the overall error on the corrected splitting (from both the peak fitting and the Ne line correction).

    Parameters
    -----------------
    pickle_str: str
        Name of Pickle with regression model for a specific part of the densimeter. Need to be in the same path as the notebook you are calling this in.

    corrected_split: pd.Series
        panda series of corrected splitting (cm-1)

    corrected_split_err: pd. Series
        panda series of error on corrected splitting (contributions from both peak fitting and the Ne correction model if relevant)

    str_d: str
            string of what density equation it came from, appended onto column headings.

    CI_split: float
        confidence interval for splitting propagation. Should be set the same as CI_

    CI_split: float

    Returns
    -----------------
    pd.DataFrame
        Dataframe of preferred density, and error from each source of uncertainty (Density_σ_dens=error from densimeter, Density_σ_split = error from spliting).

    """

    # Corrected splitting
    new_x=corrected_split
    new_x_uncertainty=corrected_split_err

    # This gets the uncertainty from the densimeter
    DiadFit_dir=Path(__file__).parent
    with open(DiadFit_dir/pickle_str, 'rb') as f:
        data = pickle.load(f)

    model = data['model']
    N_poly = model.order - 1

    Pf = data['model']
    x = data['x']
    y = data['y']

    # Calculate the residuals
    residuals = y - Pf(x)

    # Calculate the standard deviation of the residuals
    residual_std = np.std(residuals)

    # Calculate the standard errors for the new x values
    mean_x = np.nanmean(x)
    n = len(x)


    standard_errors = residual_std * np.sqrt(1 + 1 / n + (new_x - mean_x) ** 2 / np.sum((x - mean_x) ** 2))

    # Calculate the degrees of freedom
    df = len(x) - (N_poly + 1)

    # Calculate the t value for the given confidence level
    t_value_split = t.ppf((1 + CI_split) / 2, df)
    t_value_dens = t.ppf((1 + CI_dens) / 2, df)

    # Calculate the prediction intervals from the densimeter
    preferred_values = Pf(new_x)
    lower_values = preferred_values - t_value_dens * standard_errors
    upper_values = preferred_values + t_value_dens * standard_errors
    uncertainty_from_dens=(upper_values-lower_values)/2



    # Calculate the propagated uncertainty in splitting
    max_split=new_x + new_x_uncertainty
    min_split=new_x - new_x_uncertainty
    max_density=  Pf(max_split)
    min_density=  Pf(min_split)
    uncertainty_split=(max_density-min_density)/2


    # Calculate the total uncertainty in the density estimation
    total_uncertainty = np.sqrt(uncertainty_split ** 2 + uncertainty_from_dens ** 2)


    df=pd.DataFrame(data={

        str_d+'_Density': preferred_values,
        str_d + '_Density_σ': total_uncertainty,
        str_d+'_Density+1σ': preferred_values-total_uncertainty,
        str_d+'_Density-1σ': preferred_values+total_uncertainty,
        str_d+'_Density_σ_dens': (uncertainty_from_dens),
        str_d+'_Density_σ_split': (uncertainty_split),

    })

    return df
## Function for if we dont have a densimeter yet

def calculate_errors_no_densimeter(*, df_combo, Ne_pickle_str='polyfit_data.pkl',  temp='SupCrit', split_err=0, CI_split=0.67, CI_neon=0.67):
    """ This function calculates the error from just the Ne line correction method. It is largely redundant, 
    still in use for some older projects using an old workflow

    """
    time=df_combo['sec since midnight']
    Ne_corr=calculate_Ne_corr_std_err_values(pickle_str=Ne_pickle_str,
    new_x=time, CI=CI_neon)

    # Lets calculate  corrected splitting and the error on this.
    Split=df_combo['Splitting']*Ne_corr['preferred_values']
    df_combo['Corrected_Splitting']=Split
    Split_err, pk_err=propagate_error_split_neon_peakfit(Ne_corr=Ne_corr, df_fits=df_combo)
    df_combo['Corrected_Splitting_σ']=Split_err
    df_combo['Corrected_Splitting_σ_Ne']=(Ne_corr['upper_values']*df_combo['Splitting']-Ne_corr['lower_values']*df_combo['Splitting'])/2
    df_combo['Corrected_Splitting_σ_peak_fit']=pk_err

    cols_to_move = ['filename',
     'Corrected_Splitting', 'Corrected_Splitting_σ',
    'Corrected_Splitting_σ_Ne', 'Corrected_Splitting_σ_peak_fit']
    df_merge = df_combo[cols_to_move + [
        col for col in df_combo.columns if col not in cols_to_move]]

    return df_combo



def calculate_density_cornell(*, df_combo, Ne_pickle_str='polyfit_data.pkl',  temp='SupCrit', 
CI_split=0.67, CI_neon=0.67, pref_Ne=None, Ne_err=None):
    """ This function converts Diad Splitting into CO$_2$ density using the Raman at CCMR, Cornell, run by E. Gazel.
    It is currently only supported for measurements performed at 37C. (e.g. SupCrit). 
    However, if you need the lower T 24C version, we just need to hunt down the calibration data to save the densimeter pkl.  
    It fully propagates uncertainty from the densimeter, the peak fitting and the Ne correction model

    Parameters
    -------------
    df_combo: pandas DataFrame
        data frame of peak fitting information

    temp: str
        'SupCrit' if measurements done at 37C
        'RoomT' if measurements done at 24C - Not supported yet but could be added if needed. 

    CI_neon: float
        Default 0.67. Confidence interval to use, e.g. 0.67 returns 1 sigma uncertainties. If you use another number,
        note the column headings will still say sigma.

    CI_split: float
        Default 0.67. Confidence interval to use, e.g. 0.67 returns 1 sigma uncertainties. If you use another number,
        note the column headings will still say sigma.


    Either

    Ne_pickle_str: str
        Name of Ne correction model

    OR

    pref_Ne, Ne_err: float, int
        For quick and dirty fitting can pass a preferred value for your instrument before you have a chance to 
        regress the Ne lines (useful when first analysing new samples. )




    Returns
    --------------
    pd.DataFrame
        Prefered Density (based on different equatoins being merged), and intermediate calculations




    """
    df_combo_c=df_combo.copy()
    time=df_combo_c['sec since midnight']

    if Ne_pickle_str is not None:

        # Calculating the upper and lower values for Ne to get that error
        Ne_corr=calculate_Ne_corr_std_err_values(pickle_str=Ne_pickle_str,
        new_x=time, CI=CI_neon)
        # Extracting preferred correction values
        pref_Ne=Ne_corr['preferred_values']
        Split_err, pk_err=propagate_error_split_neon_peakfit(Ne_corr=Ne_corr, df_fits=df_combo_c)

        df_combo_c['Corrected_Splitting_σ']=Split_err
        df_combo_c['Corrected_Splitting_σ_Ne']=(Ne_corr['upper_values']*df_combo_c['Splitting']-Ne_corr['lower_values']*df_combo_c['Splitting'])/2
        df_combo_c['Corrected_Splitting_σ_peak_fit']=pk_err

    # If using a single value for quick dirty fitting
    else:
        Split_err, pk_err=propagate_error_split_neon_peakfit(df_fits=df_combo_c, Ne_err=Ne_err, pref_Ne=pref_Ne)


        df_combo_c['Corrected_Splitting_σ']=Split_err

        df_combo_c['Corrected_Splitting_σ_Ne']=((Ne_err+pref_Ne)*df_combo_c['Splitting']-(Ne_err-pref_Ne)*df_combo_c['Splitting'])/2
        df_combo_c['Corrected_Splitting_σ_peak_fit']=pk_err


    Split=df_combo_c['Splitting']*pref_Ne

    # This propgates the uncertainty in the splitting from peak fitting, and the Ne correction model




    if temp=='RoomT':
        raise TypeError('Sorry, this function doesnt yet support the calibration at 24C, please enter temp=SupCrit')
    if isinstance(Split, float) or isinstance(Split, int):
        Split=pd.Series(Split)
    # #if temp is "RoomT":
    DiadFit_dir=Path(__file__).parent

    LowD_RT=-38.34631 + 0.3732578*Split
    HighD_RT=-41.64784 + 0.4058777*Split- 0.1460339*(Split-104.653)**2

    # IF temp is 37
    # This gets the densimeter at low density
    pickle_str_lowr='Lowrho_polyfit_data_CMASS.pkl'
    with open(DiadFit_dir/pickle_str_lowr, 'rb') as f:
        lowrho_pickle_data = pickle.load(f)

    # This gets the densimeter at medium density
    pickle_str_medr='Mediumrho_polyfit_data_CMASS.pkl'
    with open(DiadFit_dir/pickle_str_medr, 'rb') as f:
        medrho_pickle_data = pickle.load(f)
    # This gets the densimeter at high density.
    pickle_str_highr='Highrho_polyfit_data_CMASS.pkl'
    with open(DiadFit_dir/pickle_str_highr, 'rb') as f:
        highrho_pickle_data = pickle.load(f)

    # this allocates the model
    lowrho_model = lowrho_pickle_data['model']
    medrho_model = medrho_pickle_data['model']
    highrho_model = highrho_pickle_data['model']

    # Each of these lines get the density, and then the error on that density.

    LowD_SC = pd.Series(lowrho_model(Split), index=Split.index)
    lowD_error=calculate_Densimeter_std_err_values(corrected_split=Split, corrected_split_err=Split_err,
    pickle_str=pickle_str_lowr,  CI_dens=CI_neon, CI_split=CI_split, str_d='LowD')

    MedD_SC = pd.Series(medrho_model(Split), index=Split.index)
    medD_error=calculate_Densimeter_std_err_values(corrected_split=Split, corrected_split_err=Split_err,
    pickle_str=pickle_str_medr,  CI_dens=CI_neon, CI_split=CI_split, str_d='MedD')

    HighD_SC = pd.Series(highrho_model(Split), index=Split.index)
    highD_error=calculate_Densimeter_std_err_values(corrected_split=Split, corrected_split_err=Split_err,
    pickle_str=pickle_str_highr,   CI_dens=CI_neon, CI_split=CI_split,  str_d='HighD')





    df=pd.DataFrame(data={'Preferred D': 0,
    'Corrected_Splitting': Split,
    'Preferred D_σ': 0,
    'Preferred D_σ_split': 0,
    'Preferred D_σ_Ne': 0,
    'Preferred D_σ_pkfit': 0,
     'Preferred D_σ_dens': 0,
        'in range': 'Y',
                                'Notes': 'not in range',
                                'LowD_RT':np.nan,
                                'HighD_RT': np.nan,
                                'LowD_SC': LowD_SC,
                                'LowD_SC_σ': lowD_error['LowD_Density_σ'],
                                'MedD_SC': MedD_SC,
                                'MedD_SC_σ': medD_error['MedD_Density_σ'],
                                'HighD_SC': HighD_SC,
                                'HighD_SC_σ': highD_error['HighD_Density_σ'],
                                'Temperature': temp,


                                })




    roomT=df['Temperature']=="RoomT"
    SupCrit=df['Temperature']=="SupCrit"
    # If splitting is 0
    zero=df['Corrected_Splitting']==0

    # Range for SC low density
    min_lowD_SC_Split=df['Corrected_Splitting']>=102.7623598753032
    max_lowD_SC_Split=df['Corrected_Splitting']<=103.1741034592534
    # Range for SC med density
    min_MD_SC_Split=df['Corrected_Splitting']>103.0608505403591
    max_MD_SC_Split=df['Corrected_Splitting']<=104.3836704771313
    # Range for SC high density
    min_HD_SC_Split=df['Corrected_Splitting']>=104.2538992302499
    max_HD_SC_Split=df['Corrected_Splitting']<=105.3438707618937
    # Range for Room T low density
    min_lowD_RoomT_Split=df['Corrected_Splitting']>=102.734115670188
    max_lowD_RoomT_Split=df['Corrected_Splitting']<=103.350311768435
    # Range for Room T high density
    min_HD_RoomT_Split=df['Corrected_Splitting']>=104.407308904012
    max_HD_RoomT_Split=df['Corrected_Splitting']<=105.1
    # Impossible densities, room T
    Imposs_lower_end=(df['Corrected_Splitting']>103.350311768435) & (df['Corrected_Splitting']<103.88)
    # Impossible densities, room T
    Imposs_upper_end=(df['Corrected_Splitting']<104.407308904012) & (df['Corrected_Splitting']>103.88)
    # Too low density
    Too_Low_SC=df['Corrected_Splitting']<102.7623598753032
    Too_Low_RT=df['Corrected_Splitting']<102.734115670188

    df.loc[zero, 'Preferred D']=0
    df.loc[zero, 'Notes']=0





    # If SupCrit, high density
    df.loc[ SupCrit&(min_HD_SC_Split&max_HD_SC_Split), 'Preferred D'] = HighD_SC
    df.loc[ SupCrit&(min_HD_SC_Split&max_HD_SC_Split), 'Preferred D_σ'] = highD_error['HighD_Density_σ']
    df.loc[ SupCrit&(min_HD_SC_Split&max_HD_SC_Split), 'Preferred D_σ_split'] = highD_error['HighD_Density_σ_split']
    df.loc[ SupCrit&(min_HD_SC_Split&max_HD_SC_Split), 'Preferred D_σ_dens'] = highD_error['HighD_Density_σ_dens']
    df.loc[ SupCrit&(min_HD_SC_Split&max_HD_SC_Split), 'Notes']='SupCrit, high density'
    # If SupCrit, Med density
    df.loc[SupCrit&(min_MD_SC_Split&max_MD_SC_Split), 'Preferred D'] = MedD_SC
    df.loc[SupCrit&(min_MD_SC_Split&max_MD_SC_Split), 'Preferred D_σ'] = medD_error['MedD_Density_σ']
    df.loc[SupCrit&(min_MD_SC_Split&max_MD_SC_Split), 'Preferred D_σ_split'] = medD_error['MedD_Density_σ_split']
    df.loc[SupCrit&(min_MD_SC_Split&max_MD_SC_Split), 'Preferred D_σ_dens'] = medD_error['MedD_Density_σ_dens']
    df.loc[SupCrit&(min_MD_SC_Split&max_MD_SC_Split), 'Notes']='SupCrit, Med density'

    # If SupCrit, low density
    df.loc[ SupCrit&(min_lowD_SC_Split&max_lowD_SC_Split), 'Preferred D'] = LowD_SC
    df.loc[ SupCrit&(min_lowD_SC_Split&max_lowD_SC_Split), 'Preferred D_σ'] = lowD_error['LowD_Density_σ']
    df.loc[ SupCrit&(min_lowD_SC_Split&max_lowD_SC_Split), 'Preferred D_σ_split'] = lowD_error['LowD_Density_σ_split']
    df.loc[ SupCrit&(min_lowD_SC_Split&max_lowD_SC_Split), 'Preferred D_σ_dens'] = lowD_error['LowD_Density_σ_dens']
    df.loc[SupCrit&(min_lowD_SC_Split&max_lowD_SC_Split), 'Notes']='SupCrit, low density'

    # If Supcritical, and too low
    df.loc[SupCrit&(Too_Low_SC), 'Preferred D']=LowD_SC
    df.loc[SupCrit&(Too_Low_SC), 'Notes']='Below lower calibration limit'
    df.loc[SupCrit&(Too_Low_SC), 'in range']='N'


    # If RoomT, and too low
    df.loc[roomT&(Too_Low_RT), 'Preferred D']=LowD_RT
    df.loc[roomT&(Too_Low_RT), 'Notes']='Below lower calibration limit'
    df.loc[roomT&(Too_Low_RT), 'in range']='N'

    #if splitting is zero
    SplitZero=df['Corrected_Splitting']==0
    df.loc[SupCrit&(SplitZero), 'Preferred D']=np.nan
    df.loc[SupCrit&(SplitZero), 'Notes']='Splitting=0'
    df.loc[SupCrit&(SplitZero), 'in range']='N'

    df.loc[roomT&(SplitZero), 'Preferred D']=np.nan
    df.loc[roomT&(SplitZero), 'Notes']='Splitting=0'
    df.loc[roomT&(SplitZero), 'in range']='N'


    # If impossible density, lower end
    df.loc[roomT&Imposs_lower_end, 'Preferred D'] = LowD_RT
    df.loc[roomT&Imposs_lower_end, 'Notes']='Impossible Density, low density'
    df.loc[roomT&Imposs_lower_end, 'in range']='N'

    # If impossible density, lower end
    df.loc[roomT&Imposs_upper_end, 'Preferred D'] = HighD_RT
    df.loc[roomT&Imposs_upper_end, 'Notes']='Impossible Density, high density'
    df.loc[roomT&Imposs_upper_end, 'in range']='N'


    #df.loc[zero, 'in range']='Y'
    # If high densiy, and beyond the upper calibration limit
    Upper_Cal_RT=df['Corrected_Splitting']>105.1
    Upper_Cal_SC=df['Corrected_Splitting']>105.3438707618937

    df.loc[roomT&Upper_Cal_RT, 'Preferred D'] = HighD_RT
    df.loc[roomT&Upper_Cal_RT, 'Notes']='Above upper Cali Limit'
    df.loc[roomT&Upper_Cal_RT, 'in range']='N'

    df.loc[SupCrit&Upper_Cal_SC, 'Preferred D'] = HighD_SC
    df.loc[SupCrit&Upper_Cal_SC, 'Notes']='Above upper Cali Limit'
    df.loc[SupCrit&Upper_Cal_SC, 'in range']='N'

    if Ne_pickle_str is not None:
        df_merge1=pd.concat([df_combo_c, Ne_corr], axis=1).reset_index(drop=True)
    else:
        df_merge1=df_combo_c

    df_merge=pd.concat([df, df_merge1], axis=1).reset_index(drop=True)

    df_merge = df_merge.rename(columns={'Preferred D': 'Density g/cm3'})
    df_merge = df_merge.rename(columns={'Preferred D_σ': 'σ Density g/cm3'})
    df_merge = df_merge.rename(columns={'Preferred D_σ_split': 'σ Density g/cm3 (from Ne+peakfit)'})
    df_merge = df_merge.rename(columns={'Preferred D_σ_dens': 'σ Density g/cm3 (from densimeter)'})
    df_merge = df_merge.rename(columns={'filename_x': 'filename'})


    #
    #

    if Ne_pickle_str is not None:
        cols_to_move = ['filename', 'Density g/cm3', 'σ Density g/cm3','σ Density g/cm3 (from Ne+peakfit)', 'σ Density g/cm3 (from densimeter)',
        'Corrected_Splitting', 'Corrected_Splitting_σ',
        'Corrected_Splitting_σ_Ne', 'Corrected_Splitting_σ_peak_fit', 'power (mW)', 'Spectral Center']
        df_merge = df_merge[cols_to_move + [
            col for col in df_merge.columns if col not in cols_to_move]]
    else:
        cols_to_move = ['filename', 'Density g/cm3', 'σ Density g/cm3','σ Density g/cm3 (from Ne+peakfit)', 'σ Density g/cm3 (from densimeter)',
        'Corrected_Splitting', 'Corrected_Splitting_σ',
        'Corrected_Splitting_σ_Ne', 'Corrected_Splitting_σ_peak_fit']
        df_merge = df_merge[cols_to_move + [
            col for col in df_merge.columns if col not in cols_to_move]]







    return df_merge


## functions to neatly merge secondary phases in
import os.path
from os import path
def merge_in_carb_SO2(df_combo, file1_name='Carb_Peak_fits.xlsx', file2_name='SO2_Peak_fits.xlsx', 
prefix=False, str_prefix=" ", file_ext='.txt'):
    """
    This function checks for .xlsx files with secondary phases in the path with names 
    file1_name and file2_name, if they are there it will merge them together into a dataframe.
     It then merges this with the dataframe 'df_combo' of your other fits

    Parameters
    -----------------
    df_combo: pd.DataFrame
        Dataframe of peak fitting parameters for diads
    
    file1_name: str
        Name of first excel spreadsheet of secondary phases to merge in

    file2_name: str
        Name of second excel spreadsheet of secondary phases to merge in

    prefix: bool
        If True, removes prefix followed by str_prefix from file name, 
        e.g. if your spectra are called 01 SO2_acquisition, the file name would become SO2_acquisition
        for easier merging. 
    
    file_ext: str
        Removes from file name for merging dataframes (as above).

    Returns
    ----------------
    pd.DataFrame
        Dataframe of combined peak parameters from secondary phases and diads. 

    





    """

    if path.exists(file1_name):
        Carb=pd.read_excel(file1_name)
    else:
        Carb=None
    if path.exists(file2_name):
        SO2=pd.read_excel(file2_name)
    else:
        SO2=None
    if SO2 is not None and Carb is not None:
        Sec_Phases=pd.merge(SO2, Carb, on='filename', how='outer').reset_index(drop=True)
    elif SO2 is not None and Carb is None:
        Sec_Phases=SO2
    elif SO2 is None and Carb is not None:
        Sec_Phases=Carb
    else:
        Sec_Phases=None
    if Sec_Phases is not None:
        print('Made a df!')



    # Remove these to get the pure file name
    if Sec_Phases is not None:
        file_sec_phase=extracting_filenames_generic(
            prefix=prefix, str_prefix=str_prefix, file_ext=file_ext,
            names=Sec_Phases['filename'].reset_index(drop=True))



    else:
        df_combo_sec_phase=df_combo




    if Sec_Phases is not None:
        Sec_Phases['filename']=file_sec_phase
        df_combo_sec_phase=df_combo.merge(Sec_Phases,
        on='filename', how='outer').reset_index(drop=True)



    else:
        df_combo_sec_phase=df_combo

    if 'Peak_Area_Carb' in df_combo_sec_phase.columns:
        df_combo_sec_phase['Carb_Diad_Ratio']=(df_combo_sec_phase['Peak_Area_Carb']/(df_combo_sec_phase['Diad1_Voigt_Area']
                        +df_combo_sec_phase['Diad2_Voigt_Area']))
    if 'Peak_Area_SO2' in df_combo_sec_phase.columns:
        df_combo_sec_phase['SO2_Diad_Ratio']=(df_combo_sec_phase['Peak_Area_SO2']/(df_combo_sec_phase['Diad1_Voigt_Area']
                        +df_combo_sec_phase['Diad2_Voigt_Area']))



    return df_combo_sec_phase

## Merge peak fits together
def merge_fit_files(path):
    """
    This function merges the files Discarded_df.xlsx, Weak_Diads.xlsx, Medium_Diads.xlsx, Strong_Diads.xlsx
    if they exist into one combined dataframe. 

    Parameters
    -----------------
    path: str
        Path on computer where these files are stored

    Returns
    ------------
    pd.DataFrame
        pandas dataframe where 3 sets of peak fits are merged. 

    """
    import os
    import pandas as pd

    if os.path.exists(os.path.join(path, 'Discarded_df.xlsx')):
        discard = pd.read_excel(os.path.join(path, 'Discarded_df.xlsx'))
    else:
        discard = None

    if os.path.exists(os.path.join(path, 'Weak_Diads.xlsx')):
        grp1 = pd.read_excel(os.path.join(path, 'Weak_Diads.xlsx'))
    else:
        grp1 = None

    if os.path.exists(os.path.join(path, 'Medium_Diads.xlsx')):
        grp2 = pd.read_excel(os.path.join(path, 'Medium_Diads.xlsx'))
    else:
        grp2 = None

    if os.path.exists(os.path.join(path, 'Strong_Diads.xlsx')):
        grp3 = pd.read_excel(os.path.join(path, 'Strong_Diads.xlsx'))
    else:
        grp3 = None

    df2 = pd.concat([grp1, grp2, grp3], axis=0).reset_index(drop=True)

    if discard is not None:
        discard_cols=discard[discard.columns.intersection(df2.columns)]
        df2=pd.concat([df2, discard_cols]).reset_index(drop=True)
    return df2

## New UC Berkeley using 1220

def calculate_density_ucb(*, Ne_line_combo='1117_1447', df_combo=None, temp='SupCrit', 
CI_split=0.67, CI_neon=0.67,  Ne_pickle_str=None, pref_Ne=None, Ne_err=None):
    """ This function converts Diad Splitting into CO$_2$ density using the UC Berkeley calibration line
    developed by DeVitre and Wieser in 2023. 

    Parameters
    -------------
    Ne_line_combo: str, '1117_1447', '1220_1447', '1220_1400'
        Combination of Ne lines used for drift correction

    df_combo: pandas DataFrame
        data frame of peak fitting information

    temp: str
        'SupCrit' if measurements done at 37C
        'RoomT' if measurements done at 24C - Not supported yet but could be added if needed. 

    CI_neon: float
        Default 0.67. Confidence interval to use, e.g. 0.67 returns 1 sigma uncertainties. If you use another number,
        note the column headings will still say sigma.

    CI_split: float
        Default 0.67. Confidence interval to use, e.g. 0.67 returns 1 sigma uncertainties. If you use another number,
        note the column headings will still say sigma.


    Either

    Ne_pickle_str: str
        Name of Ne correction model

    OR

    pref_Ne, Ne_err: float, int
        For quick and dirty fitting can pass a preferred value for your instrument before you have a chance to 
        regress the Ne lines (useful when first analysing new samples. )




    Returns
    --------------
    pd.DataFrame
        Prefered Density (based on different equatoins being merged), and intermediate calculations

    """
    if df_combo is not None:
        df_combo_c=df_combo.copy()
        time=df_combo_c['sec since midnight']

        if Ne_pickle_str is not None:

            # Calculating the upper and lower values for Ne to get that error
            Ne_corr=calculate_Ne_corr_std_err_values(pickle_str=Ne_pickle_str,
            new_x=time, CI=CI_neon)
            # Extracting preferred correction values
            pref_Ne=Ne_corr['preferred_values']
            Split_err, pk_err=propagate_error_split_neon_peakfit(Ne_corr=Ne_corr, df_fits=df_combo_c)

            df_combo_c['Corrected_Splitting_σ']=Split_err
            df_combo_c['Corrected_Splitting_σ_Ne']=(Ne_corr['upper_values']*df_combo_c['Splitting']-Ne_corr['lower_values']*df_combo_c['Splitting'])/2
            df_combo_c['Corrected_Splitting_σ_peak_fit']=pk_err

        # If using a single value for quick dirty fitting
        else:
            Split_err, pk_err=propagate_error_split_neon_peakfit(df_fits=df_combo_c, Ne_err=Ne_err, pref_Ne=pref_Ne)



            df_combo_c['Corrected_Splitting_σ']=Split_err

            df_combo_c['Corrected_Splitting_σ_Ne']=((Ne_err+pref_Ne)*df_combo_c['Splitting']-(Ne_err-pref_Ne)*df_combo_c['Splitting'])/2
            df_combo_c['Corrected_Splitting_σ_peak_fit']=pk_err

        Split=df_combo_c['Splitting']*pref_Ne

    else:
       Split_err=(split_err*Split).astype(float)



    # This is for if you just have splitting



    # This propgates the uncertainty in the splitting from peak fitting, and the Ne correction model




    if temp=='RoomT':
        raise TypeError('Sorry, no UC Berkeley calibration at 24C, please enter temp=SupCrit')
    if isinstance(Split, float) or isinstance(Split, int):
        Split=pd.Series(Split)
    # #if temp is "RoomT":
    DiadFit_dir=Path(__file__).parent

    LowD_RT=-38.34631 + 0.3732578*Split
    HighD_RT=-41.64784 + 0.4058777*Split- 0.1460339*(Split-104.653)**2

    # IF temp is 37
    if Ne_line_combo=='1220_1447':
    # This gets the densimeter at low density
        pickle_str_lowr='Lowrho_polyfit_dataUCB_1220_1447.pkl'
        with open(DiadFit_dir/pickle_str_lowr, 'rb') as f:
            lowrho_pickle_data = pickle.load(f)

        # This gets the densimeter at medium density
        pickle_str_medr='Mediumrho_polyfit_dataUCB_1220_1447.pkl'
        with open(DiadFit_dir/pickle_str_medr, 'rb') as f:
            medrho_pickle_data = pickle.load(f)
        # This gets the densimeter at high density.
        pickle_str_highr='Highrho_polyfit_dataUCB_1220_1447.pkl'
        with open(DiadFit_dir/pickle_str_highr, 'rb') as f:
            highrho_pickle_data = pickle.load(f)

    if Ne_line_combo=='1220_1400':
        pickle_str_lowr='Lowrho_polyfit_dataUCB_1220_1400.pkl'
        with open(DiadFit_dir/pickle_str_lowr, 'rb') as f:
            lowrho_pickle_data = pickle.load(f)

        # This gets the densimeter at medium density
        pickle_str_medr='Mediumrho_polyfit_dataUCB_1220_1400.pkl'
        with open(DiadFit_dir/pickle_str_medr, 'rb') as f:
            medrho_pickle_data = pickle.load(f)
        # This gets the densimeter at high density.
        pickle_str_highr='Highrho_polyfit_dataUCB_1220_1400.pkl'
        with open(DiadFit_dir/pickle_str_highr, 'rb') as f:
            highrho_pickle_data = pickle.load(f)

    if Ne_line_combo=='1117_1447':
    # This gets the densimeter at low density
        pickle_str_lowr='Lowrho_polyfit_data.pkl'
        with open(DiadFit_dir/pickle_str_lowr, 'rb') as f:
            lowrho_pickle_data = pickle.load(f)

        # This gets the densimeter at medium density
        pickle_str_medr='Mediumrho_polyfit_data.pkl'
        with open(DiadFit_dir/pickle_str_medr, 'rb') as f:
            medrho_pickle_data = pickle.load(f)
        # This gets the densimeter at high density.
        pickle_str_highr='Highrho_polyfit_data.pkl'
        with open(DiadFit_dir/pickle_str_highr, 'rb') as f:
            highrho_pickle_data = pickle.load(f)

    # this allocates the model
    lowrho_model = lowrho_pickle_data['model']
    medrho_model = medrho_pickle_data['model']
    highrho_model = highrho_pickle_data['model']

    # Each of these lines get the density, and then the error on that density.

    LowD_SC = pd.Series(lowrho_model(Split), index=Split.index)
    lowD_error=calculate_Densimeter_std_err_values(corrected_split=Split, corrected_split_err=Split_err,
    pickle_str=pickle_str_lowr,  CI_dens=CI_neon, CI_split=CI_split, str_d='LowD')

    MedD_SC = pd.Series(medrho_model(Split), index=Split.index)
    medD_error=calculate_Densimeter_std_err_values(corrected_split=Split, corrected_split_err=Split_err,
    pickle_str=pickle_str_medr,  CI_dens=CI_neon, CI_split=CI_split, str_d='MedD')

    HighD_SC = pd.Series(highrho_model(Split), index=Split.index)
    highD_error=calculate_Densimeter_std_err_values(corrected_split=Split, corrected_split_err=Split_err,
    pickle_str=pickle_str_highr,   CI_dens=CI_neon, CI_split=CI_split,  str_d='HighD')





    df=pd.DataFrame(data={'Preferred D': 0,
    'Corrected_Splitting': Split,
    'Preferred D_σ': 0,
    'Preferred D_σ_split': 0,
    'Preferred D_σ_Ne': 0,
    'Preferred D_σ_pkfit': 0,
     'Preferred D_σ_dens': 0,
        'in range': 'Y',
                                'Notes': 'not in range',
                                'LowD_RT':np.nan,
                                'HighD_RT': np.nan,
                                'LowD_SC': LowD_SC,
                                'LowD_SC_σ': lowD_error['LowD_Density_σ'],
                                'MedD_SC': MedD_SC,
                                'MedD_SC_σ': medD_error['MedD_Density_σ'],
                                'HighD_SC': HighD_SC,
                                'HighD_SC_σ': highD_error['HighD_Density_σ'],
                                'Temperature': temp,


                                })




    roomT=df['Temperature']=="RoomT"
    SupCrit=df['Temperature']=="SupCrit"
    # If splitting is 0
    zero=df['Corrected_Splitting']==0

    # Range for SC low density
    min_lowD_SC_Split=df['Corrected_Splitting']>=102.7623598753032
    max_lowD_SC_Split=df['Corrected_Splitting']<=103.1741034592534
    # Range for SC med density
    min_MD_SC_Split=df['Corrected_Splitting']>103.0608505403591
    max_MD_SC_Split=df['Corrected_Splitting']<=104.3836704771313
    # Range for SC high density
    min_HD_SC_Split=df['Corrected_Splitting']>=104.2538992302499
    max_HD_SC_Split=df['Corrected_Splitting']<=105.3438707618937
    # Range for Room T low density
    min_lowD_RoomT_Split=df['Corrected_Splitting']>=102.734115670188
    max_lowD_RoomT_Split=df['Corrected_Splitting']<=103.350311768435
    # Range for Room T high density
    min_HD_RoomT_Split=df['Corrected_Splitting']>=104.407308904012
    max_HD_RoomT_Split=df['Corrected_Splitting']<=105.1
    # Impossible densities, room T
    Imposs_lower_end=(df['Corrected_Splitting']>103.350311768435) & (df['Corrected_Splitting']<103.88)
    # Impossible densities, room T
    Imposs_upper_end=(df['Corrected_Splitting']<104.407308904012) & (df['Corrected_Splitting']>103.88)
    # Too low density
    Too_Low_SC=df['Corrected_Splitting']<102.7623598753032
    Too_Low_RT=df['Corrected_Splitting']<102.734115670188

    df.loc[zero, 'Preferred D']=0
    df.loc[zero, 'Notes']=0





    # If SupCrit, high density
    df.loc[ SupCrit&(min_HD_SC_Split&max_HD_SC_Split), 'Preferred D'] = HighD_SC
    df.loc[ SupCrit&(min_HD_SC_Split&max_HD_SC_Split), 'Preferred D_σ'] = highD_error['HighD_Density_σ']
    df.loc[ SupCrit&(min_HD_SC_Split&max_HD_SC_Split), 'Preferred D_σ_split'] = highD_error['HighD_Density_σ_split']
    df.loc[ SupCrit&(min_HD_SC_Split&max_HD_SC_Split), 'Preferred D_σ_dens'] = highD_error['HighD_Density_σ_dens']
    df.loc[ SupCrit&(min_HD_SC_Split&max_HD_SC_Split), 'Notes']='SupCrit, high density'
    # If SupCrit, Med density
    df.loc[SupCrit&(min_MD_SC_Split&max_MD_SC_Split), 'Preferred D'] = MedD_SC
    df.loc[SupCrit&(min_MD_SC_Split&max_MD_SC_Split), 'Preferred D_σ'] = medD_error['MedD_Density_σ']
    df.loc[SupCrit&(min_MD_SC_Split&max_MD_SC_Split), 'Preferred D_σ_split'] = medD_error['MedD_Density_σ_split']
    df.loc[SupCrit&(min_MD_SC_Split&max_MD_SC_Split), 'Preferred D_σ_dens'] = medD_error['MedD_Density_σ_dens']
    df.loc[SupCrit&(min_MD_SC_Split&max_MD_SC_Split), 'Notes']='SupCrit, Med density'

    # If SupCrit, low density
    df.loc[ SupCrit&(min_lowD_SC_Split&max_lowD_SC_Split), 'Preferred D'] = LowD_SC
    df.loc[ SupCrit&(min_lowD_SC_Split&max_lowD_SC_Split), 'Preferred D_σ'] = lowD_error['LowD_Density_σ']
    df.loc[ SupCrit&(min_lowD_SC_Split&max_lowD_SC_Split), 'Preferred D_σ_split'] = lowD_error['LowD_Density_σ_split']
    df.loc[ SupCrit&(min_lowD_SC_Split&max_lowD_SC_Split), 'Preferred D_σ_dens'] = lowD_error['LowD_Density_σ_dens']
    df.loc[SupCrit&(min_lowD_SC_Split&max_lowD_SC_Split), 'Notes']='SupCrit, low density'

    # If Supcritical, and too low
    df.loc[SupCrit&(Too_Low_SC), 'Preferred D']=LowD_SC
    df.loc[SupCrit&(Too_Low_SC), 'Notes']='Below lower calibration limit'
    df.loc[SupCrit&(Too_Low_SC), 'in range']='N'


    # If RoomT, and too low
    df.loc[roomT&(Too_Low_RT), 'Preferred D']=LowD_RT
    df.loc[roomT&(Too_Low_RT), 'Notes']='Below lower calibration limit'
    df.loc[roomT&(Too_Low_RT), 'in range']='N'

    #if splitting is zero
    SplitZero=df['Corrected_Splitting']==0
    df.loc[SupCrit&(SplitZero), 'Preferred D']=np.nan
    df.loc[SupCrit&(SplitZero), 'Notes']='Splitting=0'
    df.loc[SupCrit&(SplitZero), 'in range']='N'

    df.loc[roomT&(SplitZero), 'Preferred D']=np.nan
    df.loc[roomT&(SplitZero), 'Notes']='Splitting=0'
    df.loc[roomT&(SplitZero), 'in range']='N'


    # If impossible density, lower end
    df.loc[roomT&Imposs_lower_end, 'Preferred D'] = LowD_RT
    df.loc[roomT&Imposs_lower_end, 'Notes']='Impossible Density, low density'
    df.loc[roomT&Imposs_lower_end, 'in range']='N'

    # If impossible density, lower end
    df.loc[roomT&Imposs_upper_end, 'Preferred D'] = HighD_RT
    df.loc[roomT&Imposs_upper_end, 'Notes']='Impossible Density, high density'
    df.loc[roomT&Imposs_upper_end, 'in range']='N'


    #df.loc[zero, 'in range']='Y'
    # If high densiy, and beyond the upper calibration limit
    Upper_Cal_RT=df['Corrected_Splitting']>105.1
    Upper_Cal_SC=df['Corrected_Splitting']>105.3438707618937

    df.loc[roomT&Upper_Cal_RT, 'Preferred D'] = HighD_RT
    df.loc[roomT&Upper_Cal_RT, 'Notes']='Above upper Cali Limit'
    df.loc[roomT&Upper_Cal_RT, 'in range']='N'

    df.loc[SupCrit&Upper_Cal_SC, 'Preferred D'] = HighD_SC
    df.loc[SupCrit&Upper_Cal_SC, 'Notes']='Above upper Cali Limit'
    df.loc[SupCrit&Upper_Cal_SC, 'in range']='N'

    if Ne_pickle_str is not None:
        df_merge1=pd.concat([df_combo_c, Ne_corr], axis=1).reset_index(drop=True)
    else:
        df_merge1=df

    df_merge=pd.concat([df, df_merge1], axis=1).reset_index(drop=True)

    df_merge = df_merge.rename(columns={'Preferred D': 'Density g/cm3'})
    df_merge = df_merge.rename(columns={'Preferred D_σ': 'σ Density g/cm3'})
    df_merge = df_merge.rename(columns={'Preferred D_σ_split': 'σ Density g/cm3 (from Ne+peakfit)'})
    df_merge = df_merge.rename(columns={'Preferred D_σ_dens': 'σ Density g/cm3 (from densimeter)'})
    df_merge = df_merge.rename(columns={'filename_x': 'filename'})


    #
    #

    if Ne_pickle_str is not None:
        cols_to_move = ['filename', 'Density g/cm3', 'σ Density g/cm3','σ Density g/cm3 (from Ne+peakfit)', 'σ Density g/cm3 (from densimeter)',
        'Corrected_Splitting', 'Corrected_Splitting_σ',
        'Corrected_Splitting_σ_Ne', 'Corrected_Splitting_σ_peak_fit', 'power (mW)', 'Spectral Center']
        df_merge = df_merge[cols_to_move + [
            col for col in df_merge.columns if col not in cols_to_move]]
    elif pref_Ne is not None:
        cols_to_move = ['filename', 'Density g/cm3', 'σ Density g/cm3','σ Density g/cm3 (from Ne+peakfit)', 'σ Density g/cm3 (from densimeter)',
        'Corrected_Splitting', 'Corrected_Splitting_σ',
        'Corrected_Splitting_σ_Ne', 'Corrected_Splitting_σ_peak_fit']
        df_merge = df_merge[cols_to_move + [
            col for col in df_merge.columns if col not in cols_to_move]]








    return df_merge





import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import patches
import lmfit
from lmfit.models import GaussianModel, VoigtModel, LinearModel, ConstantModel
from scipy.signal import find_peaks
import os
import re
from os import listdir
from os.path import isfile, join
import pickle
from scipy import stats


encode="ISO-8859-1"


def calculate_generic_std_err_values(*, pickle_str, new_x, CI=0.67):

    """
    This function loads a model from a pickle file and calculates standard error values based on the model residuals, considering a confidence interval.

    Parameters
    -----------------
    pickle_str (str)
        The path to the pickle file containing the model.
    new_x (np.array):
        x values for calculation.
    CI (float, optional):
        The confidence level for prediction intervals (default: 0.67).

    Returns:
    - pandas.DataFrame

    """
    # Load the model and the data from the pickle file
    with open(pickle_str, 'rb') as f:
        data = pickle.load(f)

    model = data['model']
    N_poly = model.order - 1
    Pf = data['model']
    x = data['x']
    y = data['y']

    # Calculate the residuals
    residuals = y - Pf(x)

    # Calculate the standard deviation of the residuals
    residual_std = np.std(residuals)

    # Calculate the standard errors for the new x values
    mean_x = np.mean(x)
    n = len(x)
    standard_errors = residual_std * np.sqrt(1 + 1/n + (new_x - mean_x)**2 / np.sum((x - mean_x)**2))

    # Calculate the degrees of freedom
    df = len(x) - (N_poly + 1)

    # Calculate the t value for the given confidence level
    t_value = stats.t.ppf((1 + CI) / 2, df)

    # Calculate the prediction intervals
    preferred_values = Pf(new_x)
    lower_values = preferred_values - t_value * standard_errors
    upper_values = preferred_values + t_value * standard_errors

    df=pd.DataFrame(data={
        'time': new_x,
        'preferred_values': preferred_values,
        'lower_values': lower_values,
        'upper_values': upper_values
    })

    return df

def plot_and_save_CO2cali_pickle(*, cali_data, CO2_dens_col='rho',Split_col='Split', split_error='split_err',CO2_dens_error='dens_err', density_range, N_poly=3, CI=0.67, std_error=True, save_fig=False,eq_division='ccmr',save_suffix=''):
    """
    This function calculates and saves three polynomial regression models as pickles for CO2 calibration data according to DeVitre et al. 2021 (low density, mid density and high density).
    The models are saved as pickles and the plot is saved as well if desired.

    Parameters
    -----------------
    cali_data (pandas.DataFrame):
        A DataFrame containing calibration data (should at least contain a density and a fermi splitting column)

    CO2_dens_col (str, optional):
        The column name corresponding to CO2 density in the cali_data DataFrame (default: 'rho').

    Split_col (str, optional):
        The column name corresponding to Fermi splitting values in the cali_data DataFrame  (default: 'Split').

    split_error (str, float, or array-like, optional):
        The column name corresponding to error in fermi splitting OR an array of splitting errors OR a float (default: 'split_err').

    CO2_dens_error (str, float or array-like, optional):
        The column name corresponding to error in CO2 density OR an array of density errors OR a float(default: 'dens_err').

    density_range (str):
        The density range to be fit ('Low', 'Medium', or 'High').

    N_poly (int, optional):
        The degree of the polynomial fit (default: 3).

    CI (float, optional):
        The confidence level for prediction intervals (default: 0.67).

    std_error (bool, optional):
        Whether to calculate and plot standard error (default: True).

    save_fig (bool, optional):
        Whether to save the plot as an image (default: False).

    eq_division (str, optional):
        Method for dividing the data based on density ('ccmr' or 'cmass', default: 'ccmr'). CCMR corresponds to the limits for each section as shown in DeVitre et al., 2021 (Chem. Geo), cmass is for those in DeVitre et al., 2023 (J. Volcanica)

    save_suffix (str, optional):
        Suffix to be added to the saved file names (default: '').

    Returns:
        Saves a .pkl file, doesnt return anything.

    """
# Define the x and y values
    try:
        if eq_division=='ccmr':
            lowcut=0.17
            midcut_low=0.12
            midcut_high=0.72
            highcut=0.65
        elif eq_division=='cmass':
            lowcut=0.20
            midcut_low=0.13
            midcut_high=0.70
            highcut=0.65
        if density_range == 'Low':
            cali_data = cali_data[cali_data[CO2_dens_col] < lowcut]

            prefix='Lowrho_'
        elif density_range == 'Medium':
            cali_data = cali_data[cali_data[CO2_dens_col].between(midcut_low, midcut_high)]

            prefix='Mediumrho_'
        elif density_range == 'High':
            cali_data = cali_data[cali_data[CO2_dens_col] > highcut]

            prefix='Highrho_'
        else:
            raise ValueError("Invalid density range. Please choose 'Low', 'Medium', or 'High'.")
    except ValueError as e:
        print(f"Warning: {e}")
        return

    x_all   = cali_data[Split_col].values
    y_all = cali_data[CO2_dens_col].values

    if not isinstance(split_error,str)==True:
        x_err=pd.Series(split_error, index=cali_data[CO2_dens_col].index).values
    else:
        x_err=cali_data[split_error].values

    if not isinstance(CO2_dens_error,str)==True:
        y_err=pd.Series(CO2_dens_error, index=cali_data[CO2_dens_col].index).values
    else:
        y_err=cali_data[CO2_dens_error].values

    non_nan_indices = ~np.isnan(x_all) & ~np.isnan(y_all)

    # Filter out NaN values
    x = x_all[non_nan_indices]
    y = y_all[non_nan_indices]
    # Perform polynomial regression

    coefficients = np.polyfit(x, y, N_poly)
    Pf = np.poly1d(coefficients)


    # Save the model and the data to a pickle file
    data = {'model': Pf, 'x': x, 'y': y}
    with open(prefix+'polyfit_data'+save_suffix+'.pkl', 'wb') as f:
        pickle.dump(data, f)

    if std_error is True:
        new_x_plot = np.linspace(np.min(x), np.max(x), 100)
        new_calidf = calculate_generic_std_err_values(pickle_str=prefix+'polyfit_data'+save_suffix+'.pkl',
                                                    new_x=pd.Series(new_x_plot), CI=CI)
        # Calculate R-squared and p-value
        residuals = y - Pf(x)
        ssr = np.sum(residuals**2)
        sst = np.sum((y - np.mean(y))**2)
        r_squared = 1 - (ssr / sst)
        df = len(x) - (N_poly + 1)
        p_value = 1 - stats.f.cdf(r_squared * (len(x) - 1), N_poly, df)
        legend_label = f'best fit: R$^2$ = {r_squared:.4f}, p-val = {p_value:.1e}'
    else:
        legend_label = 'best fit'


    # Now lets plot the confidence interval
    fig, (ax1) = plt.subplots(1, 1, figsize=(10,5))
    ax1.errorbar(x, y, xerr=x_err,yerr=y_err,
            fmt='o', ecolor='grey', elinewidth=0.8, mfc='grey', ms=5, mec='k',capsize=3,barsabove=True)

    ax1.plot(new_x_plot, new_calidf['preferred_values'], '-k', label=legend_label)
    # ax1.plot(new_x_plot, new_calidf['lower_values'], ':k', label='lower vals')
    # ax1.plot(new_x_plot, new_calidf['upper_values'], ':k', label='upper vals')
    ax1.fill_between(new_x_plot, new_calidf['lower_values'], new_calidf['upper_values'], color='gray', alpha=0.2,zorder=-2,label='Prediction interval')
    ax1.set_xlabel('Split')
    ax1.set_ylabel('Density')
    ax1.legend()

    ax1.ticklabel_format(useOffset=False)
    # this sets the ordinal suffix for the polynomial degree in the title
    if N_poly == 1:
        suffix = 'st'
    elif N_poly == 2:
        suffix = 'nd'
    elif N_poly == 3:
        suffix = 'rd'
    else:
        suffix='th'

    ax1.set_title(f"{N_poly}$^{{{suffix}}}$ degree polynomial: "+str(100*CI) + ' % prediction interval')


    if save_fig is True:
        fig.savefig(prefix+'cali_line'+save_suffix+'.png')


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import math
from scipy.optimize import newton
import warnings

from DiadFit.density_depth_crustal_profiles import *
from DiadFit.CO2_EOS import *

## Functions to find P when the user chooses to start with a depth. It requires input of a crustal model

class config_crustalmodel:
    """
    A configuration class for specifying parameters of the crustal model. 

    Attributes:
    - crust_dens_kgm3 (float): The density of the crust in kilograms per cubic meter (kg/m^3).
    - d1 (float): The depth boundary for the first layer in kilometers (km).
    - d2 (float): The depth boundary for the second layer in kilometers (km).
    - rho1 (float): The density of the first layer in kilograms per cubic meter (kg/m^3).
    - rho2 (float): The density of the second layer in kilograms per cubic meter (kg/m^3).
    - rho3 (float): The density of the third layer in kilograms per cubic meter (kg/m^3).
    - model (str): The name of the model used for crustal calculations.
    """
    def __init__(self, crust_dens_kgm3=None,
                 d1=None, d2=None, rho1=None, rho2=None, rho3=None, model=None):
        self.crust_dens_kgm3 = crust_dens_kgm3
        self.d1 = d1
        self.d2 = d2
        self.rho1 = rho1
        self.rho2 = rho2
        self.rho3 = rho3
        self.model = model

def objective_function_depth(P_kbar, target_depth_km, crust_dens_kgm3,
                            d1, d2, rho1, rho2, rho3, model):
    """
    Calculate the difference between the current depth and the target depth
    given pressure (P_kbar) and other parameters.

    Parameters:
    - P_kbar (float): The pressure in kilobars (kbar) to be used in the depth calculation.
    - target_depth_km (float): The desired depth in kilometers (km).
    - crust_dens_kgm3 (float): The density of the crust in kilograms per cubic meter (kg/m^3).
    - d1, d2 (float): Depth boundaries for different layers (km).
    - rho1, rho2, rho3 (float): Densities for different layers (kg/m^3).
    - model (str): The name of the model used for the depth calculation.

    Returns:
    - float: The difference between the current depth and the target depth.
    """

    current_depth = convert_pressure_to_depth(P_kbar=P_kbar, crust_dens_kgm3=crust_dens_kgm3, g=9.81,
                                              d1=d1, d2=d2, rho1=rho1, rho2=rho2, rho3=rho3, model=model)[0]

    return current_depth - target_depth_km

def find_P_for_kmdepth(target_depth_km, crustal_model_config=config_crustalmodel(), initial_P_guess_kbar=0, tolerance=0.1):
    """
    Approximate the pressure (P_kbar) based on the target depth using the Newton-Raphson method.

    Parameters:
    - target_depth_km (float, pd.Series, list): The desired depth(s) in kilometers (km).
    - initial_P_guess_kbar (float, optional): Initial guess for the pressure in kilobars (kbar). Default is 0.
    - crustal_model_config (object, optional): Configuration object containing crustal model parameters.
        - crust_dens_kgm3 (float, optional): The density of the crust in kilograms per cubic meter (kg/m^3). Default is None.
        - d1, d2 (float, optional): Depth boundaries for different layers (km). Default is None.
        - rho1, rho2, rho3 (float, optional): Densities for different layers (kg/m^3). Default is None.
        - model (str, optional): The name of the model used for depth calculation. Default is None.
    - tolerance (float, optional): Tolerance for the Newton-Raphson method. The pressure estimate should be within this tolerance of the true value. Default is 0.1.

    Returns:
    - float or pd.Series or list: The estimated pressure(s) (P_kbar) that correspond to the target depth(s).
    
    Notes:
    - If the target_depth_km is a single value, a float is returned.
    - If the target_depth_km is a Pandas Series, a Pandas Series is returned.
    - If the target_depth_km is a list, a list of floats is returned.

    If crustal parameters are not provided in the crustal_model_config object, a single step model with a crustal density = 2750 kg/cm3 will be used.

    """

    if isinstance(target_depth_km, (float, int)):
        target_depth_km = [target_depth_km]  

    pressures = []

    for depth in target_depth_km:
        if all(v is None for v in [crustal_model_config.crust_dens_kgm3, crustal_model_config.d1, crustal_model_config.d2, crustal_model_config.rho1, crustal_model_config.rho2, crustal_model_config.rho3, crustal_model_config.model]):
            crustal_model_config.crust_dens_kgm3 = 2750
            warning_message = "\033[91mNo crustal parameters were provided, setting crust_dens_kgm3 to 2750. \nPlease use config_crustalmodel(...) to set your desired crustal model parameters.\033[0m"
            warnings.simplefilter("always")
            warnings.warn(warning_message, Warning, stacklevel=2)
        
        # Use the Newton-Raphson method for each target depth
        pressure = newton(objective_function_depth, initial_P_guess_kbar, args=(depth, crustal_model_config.crust_dens_kgm3, crustal_model_config.d1, crustal_model_config.d2, crustal_model_config.rho1, crustal_model_config.rho2, crustal_model_config.rho3, crustal_model_config.model), tol=tolerance)
        pressures.append(pressure)

    if isinstance(target_depth_km, (float, int)):
        return pressures[0]
    elif isinstance(target_depth_km, pd.Series):
        return pd.Series(pressures)
    else:
        return pressures

## Auxilliary functions for the stretching models

# Calculate decompression steps for polybaric model (Pressure, Depth, dt)

def calculate_DPdt(ascent_rate_ms,crustal_model_config=config_crustalmodel(),D_initial_km=None,D_final_km=None,D_step=100,initial_P_guess_kbar=0, tolerance=0.001):
    """
    Calculate the decompression rate (dP/dt) during ascent.

    Parameters:
    - ascent_rate_ms (float): Ascent rate in meters per second.
    - D_initial_km (float, optional): Initial depth in kilometers. Default is 30 km.
    - D_final_km (float, optional): Final depth in kilometers. Default is 0 km.
    - D_step (int, optional): Number of depth steps for calculation. Default is 100.
    - initial_P_guess_kbar (float, optional): Initial guess for pressure in kilobars (kbar). Default is 0.
    - tolerance (float, optional): Tolerance for pressure estimation. Default is 0.001.

    Returns:
    - D (pd.Series): Depth values in kilometers.
    - Pexternal_steps_MPa (list): Lithostatic pressure values in megapascals (MPa) at each depth step.
    - dt (float): Time step for the integration.
    """

    if D_initial_km is None or D_final_km is None or D_initial_km <= D_final_km:
        raise ValueError("Both D_initial_km and D_final_km must be provided, and D_initial_km must be larger than D_final_km")
    if D_initial_km>30 and D_step <= 80 and ascent_rate_ms <= 0.02:
        raise Warning("Your D_step is too small, the minimum recommended for ascent rates below 0.02 m/s is 80")
    D = pd.Series(list(np.linspace(D_initial_km, D_final_km, D_step)))  # km

    Pexternal_steps=find_P_for_kmdepth(D, crustal_model_config=crustal_model_config, initial_P_guess_kbar=initial_P_guess_kbar, tolerance=tolerance)
    Pexternal_steps_MPa=Pexternal_steps*100

    # Time steps of the ascent
    ascent_rate = ascent_rate_ms / 1000  # km/s
    D_change = abs(D.diff())
    time_series = D_change / ascent_rate  # calculates the time in between each step based on ascent rate
    dt_s = time_series.max()  # this sets the time step for the iterations later

    return D, Pexternal_steps_MPa, dt_s

# Olivine creep constants
class power_creep_law_constants:
    """
    Olivine power-law creep constants used in the stretching model (Wanamaker and Evans, 1989).

    Attributes:
    - A (float): Creep law constant A (default: 3.9e3).
    - n (float): Creep law constant n (default: 3.6).
    - Q (float): Activation energy for dislocation motions in J/mol (default: 523000).
    - IgasR (float): Gas constant in J/(mol*K) (default: 8.314).
    """
    def __init__(self):
        self.A = 3.9*10**3 #7.0 * 10**4
        self.n = 3.6 #3
        self.Q = 523000 # 520 Activation energy for dislocation motions in J/mol
        self.IgasR= 8.314  # Gas constant in J/(mol*K)

# Helper function to calculate change in radius over time (dR/dt)
def calculate_dR_dt(*,R_m, b_m, T_K,  Pinternal_MPa, Pexternal_MPa):
    """
    Calculate the rate of change of inclusion radius (dR/dt) based on power law creep.

    Parameters:
    - R_m (float): Inclusion radius in meters.
    - b_m (float): Distance to the crystal defect structures. Wanamaker and Evans (1989) use R/b=1/1000.
    - T_K (float): Temperature in Kelvin.
    - Pinternal_MPa (float): Internal pressure in MPa.
    - Pexternal_MPa (float): External pressure in MPa.

    Returns:
    - dR_dt (float): Rate of change of inclusion radius in meters per second.
    """
    
    pl_Cs = power_creep_law_constants()
    if Pinternal_MPa<Pexternal_MPa==True:
        S=-1
    else:
        S=1
    try:
        dR_dt = 2 * (S * pl_Cs.A * math.exp(-pl_Cs.Q / (pl_Cs.IgasR * T_K))) * (((R_m * b_m)**3) / (((b_m**(3 / pl_Cs.n)) - (R_m**(3 / pl_Cs.n))))**pl_Cs.n) * (((3 * abs(Pinternal_MPa - Pexternal_MPa)) / (2 * pl_Cs.n))**pl_Cs.n) / R_m**2
        return dR_dt

    except FloatingPointError:
        return np.nan
    
# Helper function to numerically solve for R (uses Runge-Kutta method, orders 1-4)   
def get_R(R_m,b_m,T_K,Pinternal_MPa,Pexternal_MPa,dt_s,method='RK1'):
    """
    Find the radius R of an FI over a time step using the Runge-Kutta numerical method. 
    Options are order 1 to 4 RK methods, such as RK1 (Euler), RK2 (Heun), RK3, or RK4.

    Parameters:
    - R_m (float): Initial FI Radius in meters.
    - b_m (float): Distance to defect structures in meters.
    - T_K (float): Temperature in Kelvin.
    - Pinternal_MPa (float): Internal pressure in MPa.
    - Pexternal_MPa (float): External pressure in MPa.
    - dt_s (float): The time step for integration in seconds.
    - method (str, optional): The numerical integration method to use. Default is 'RK1'.

    Returns:
    - tuple: A tuple containing the updated value of R_m and the derivative dR_dt (rate of change of R).
    """
    if method == 'RK1'or 'Euler':
        k1 = dt_s * calculate_dR_dt(R_m=R_m, b_m=b_m, T_K=T_K, Pinternal_MPa=Pinternal_MPa, Pexternal_MPa=Pexternal_MPa)
        dR=k1
        dR_dt = dR / dt_s
        R_m += dR
    elif method == 'RK2' or 'Heun':
        k1 = dt_s * calculate_dR_dt(R_m=R_m, b_m=b_m, T_K=T_K, Pinternal_MPa=Pinternal_MPa, Pexternal_MPa=Pexternal_MPa)
        k2 = dt_s * calculate_dR_dt(R_m=R_m + 0.5 * k1, b_m=b_m, T_K=T_K, Pinternal_MPa=Pinternal_MPa, Pexternal_MPa=Pexternal_MPa)
        dR = ((k1 + k2) / 2)
        dR_dt = dR / dt_s
        R_m += dR
    elif method == 'RK3':
        k1 = dt_s * calculate_dR_dt(R_m=R_m, b_m=b_m, T_K=T_K, Pinternal_MPa=Pinternal_MPa, Pexternal_MPa=Pexternal_MPa)
        k2 = dt_s * calculate_dR_dt(R_m=R_m + 0.5 * k1, b_m=b_m, T_K=T_K, Pinternal_MPa=Pinternal_MPa, Pexternal_MPa=Pexternal_MPa)
        k3 = dt_s * calculate_dR_dt(R_m=R_m + 0.5 * k2, b_m=b_m, T_K=T_K, Pinternal_MPa=Pinternal_MPa, Pexternal_MPa=Pexternal_MPa)
        dR = ((k1 + 4 * k2 + k3) / 6)
        dR_dt = dR / dt_s
        R_m += dR
    elif method == 'RK4':
        k1 = dt_s * calculate_dR_dt(R_m=R_m, b_m=b_m, T_K=T_K, Pinternal_MPa=Pinternal_MPa, Pexternal_MPa=Pexternal_MPa)
        k2 = dt_s * calculate_dR_dt(R_m=R_m + 0.5 * k1, b_m=b_m, T_K=T_K, Pinternal_MPa=Pinternal_MPa, Pexternal_MPa=Pexternal_MPa)
        k3 = dt_s * calculate_dR_dt(R_m=R_m + 0.5 * k2, b_m=b_m, T_K=T_K, Pinternal_MPa=Pinternal_MPa, Pexternal_MPa=Pexternal_MPa)
        k4 = dt_s * calculate_dR_dt(R_m=R_m + k3, b_m=b_m, T_K=T_K, Pinternal_MPa=Pinternal_MPa, Pexternal_MPa=Pexternal_MPa)
        dR = ((k1 + 2 * k2 + 2 * k3 + k4) / 6)
        dR_dt = dR / dt_s
        R_m += dR / 6
    else:
        raise ValueError("Unsupported numerical method. Choose from 'RK1' or 'Euler', 'RK2' or 'Huen', 'RK3', 'RK4'")

    return R_m, dR_dt

## Functions to calculate P, CO2dens, CO2mass and V

# Calculate initial CO2 density in g/cm3 and CO2 mass in g

def get_initial_CO2(R_m, T_K, P_MPa, EOS='SW96', return_volume=False):
    """
    Calculate the initial density and mass of CO2 inside a fluid inclusion (FI).

    Parameters:
    - R_m (float): The radius of the fluid inclusion (FI), in meters.
    - T_K (float): The temperature, in Kelvin.
    - P_MPa (float): The pressure, in MegaPascals (MPa).
    - EOS (str, optional): The equation of state (EOS) to use for density calculations.
        Can be one of: 'ideal' (ideal gas), 'SW96' (Span and Wagner EOS 1996), or 'SP94' (Sterner and Pitzer EOS 1994).
        Defaults to 'SW96'.
    - return_volume (bool, optional): Whether to return the volume of the FI along with density and mass. Defaults to False.

    Returns:
    - tuple or float: If return_volume is True, returns a tuple containing (V, CO2_dens_initial, CO2_mass_initial), where:
      - V (float): The volume of the fluid inclusion (FI), in cubic meters (m³).
      - CO2_dens_initial (float): The initial density of CO2 within the FI, in grams per cubic centimeter (g/cm³).
      - CO2_mass_initial (float): The initial mass of CO2 within the FI, in grams (g).

    - If return_volume is False, returns a tuple containing (CO2_dens_initial, CO2_mass_initial).

    """

    valid_EOS = ['ideal', 'SW96', 'SP94']

    try:
        if EOS not in valid_EOS:
            raise ValueError("EOS can only be 'ideal', 'SW96', or 'SP94'")
        
        if EOS == 'ideal':
            R_gas = 8.314  # J.mol/K J: kg·m²/s²
            V = 4/3 * math.pi * R_m**3  # m3
            P = P_MPa * 10**6  # convert MPa to Pa
            M = 44.01 / 1000  # kg/mol

            CO2_mass_kg = P * V * M / (R_gas * T_K)  # CO2 mass in kg
            rho = (CO2_mass_kg / V)  # rho in kg/m3

            CO2_dens_initial = rho / 1000  # CO2 density in g/cm3
            CO2_mass_initial = CO2_mass_kg / 1000  # CO2 mass in g

        else:
            R_m = R_m * 10**2  # radius in cm
            V = 4/3 * math.pi * R_m**3  # cm3, Volume of the FI, assume sphere
            P_kbar = P_MPa / 100  # Internal pressure of the FI, convert to kbar

            CO2_dens_initial = calculate_rho_for_P_T(EOS=EOS, P_kbar=P_kbar, T_K=T_K)[0]  # CO2 density in g/cm3
            CO2_mass_initial = CO2_dens_initial * V  # CO2 mass in g

        if return_volume:
            return V, CO2_dens_initial, CO2_mass_initial
        else:
            return CO2_dens_initial, CO2_mass_initial

    except ValueError as ve:
        raise ve

# Calculate CO2 density in g/cm3 and P in MPa for fixed CO2 mass in g

def get_CO2dens_P(R_m,T_K,CO2_mass,EOS='SW96',return_volume=False):
    """
    Calculate the density and pressure of CO2 inside a fluid inclusion (FI).

    Parameters:
    - R_m (float): The radius of the fluid inclusion (FI), in meters.
    - T_K (float): The temperature, in Kelvin.
    - CO2_mass (float): The mass of CO2 within the FI, in grams (g).
    - EOS (str, optional): The equation of state (EOS) to use for density and pressure calculations.
        Can be one of: 'ideal' (ideal gas), 'SW96' (Span and Wagner 1996), or 'SP94' (Sterner and Pitzer 1994).
        Defaults to 'SW96'.
    - return_volume (bool, optional): Whether to return the volume of the FI along with density and pressure. Defaults to False.

    Returns:
    - tuple or float: If return_volume is True, returns a tuple containing (V, CO2_dens, P), where:
      - V (float): The volume of the fluid inclusion (FI), in cubic meters (m³).
      - CO2_dens (float): The density of CO2 within the FI, in grams per cubic centimeter (g/cm³).
      - P (float): The pressure of CO2 within the FI, in MegaPascals (MPa).

    - If return_volume is False, returns a tuple containing (CO2_dens, P).

    """
    valid_EOS = ['ideal', 'SW96', 'SP94']

    try:
        if EOS not in valid_EOS:
            raise ValueError("EOS can only be 'ideal', 'SW96', or 'SP94'")
        
        if EOS == 'ideal':
            R_gas = 8.314  # J.mol/K J: kg·m²/s²
            V = 4/3 * math.pi * R_m**3  # m3
            M = 44.01 / 1000  # kg/mol

            CO2_mass_kg=CO2_mass*1000
            P=CO2_mass_kg*R_gas*T_K/(M*V) #P in Pa
            CO2_dens=(CO2_mass_kg/V) # CO2 density in kg/m3

            P=P/(10**6) #P in MPa
            CO2_dens=CO2_dens/1000 #rho in g/cm3

        else:
            R_m=R_m*10**2 #FI radius, convert to cm
            V=4/3*math.pi*R_m**3 #cm3, Volume of the FI, assume sphere

            CO2_dens=CO2_mass/V # CO2 density in g/cm3

            try:
                P=calculate_P_for_rho_T(EOS=EOS,CO2_dens_gcm3=CO2_dens, T_K=T_K)['P_MPa'][0] #g/cm3, CO2 density

            except ValueError:
                P=np.nan
    
        if return_volume:
            return V, CO2_dens, P
        else:
            return CO2_dens,P

    except ValueError as ve:
        raise ve
    
## Stretching Models

# This function is to model FI stretching during decompression and ascent 
def stretch_in_ascent(*, R_m, b_m, T_K, ascent_rate_ms, depth_path_ini_fin_step=[100, 0, 100],
                      crustal_model_config=config_crustalmodel(crust_dens_kgm3=2750),
                      EOS, plotfig=True, report_results='fullpath',
                      initial_P_guess_kbar=0, tolerance=0.001,method='RK4',update_b=False):
    """
    Simulate the stretching of a CO2-dominated fluid inclusion (FI) during ascent.

    Parameters:
    - R_m (float): The initial radius of the fluid inclusion (FI), in meters.
    - b_m (float): The initial distance to a crystal defect (rim, crack, etc) from the FI center, in meters.
    - T_K (float): The temperature, in Kelvin.
    - ascent_rate_ms (float): The ascent rate, in meters per second (m/s).
    - depth_path_ini_fin_step (list, optional): A list containing [initial_depth_km, final_depth_km, depth_step].
      Defaults to [100, 0, 100], representing the depth path from initial to final depth in a number of steps.
    - crustal_model_config (dict, optional): Configuration parameters for the crustal model.
      Defaults to a predefined configuration with a crustal density of 2750 kg/m³.
    - EOS (str): The equation of state (EOS) to use for density calculations. Can be one of: 'ideal' (ideal gas),
      'SW96' (Span and Wagner 1996), or 'SP94' (Sterner and Pitzer 1994).
    - plotfig (bool, optional): Whether to plot figures showing the changes in depth and CO2 density. Defaults to True.
    - report_results (str, optional): The type of results to report. Can be 'fullpath', 'startendonly', or 'endonly'.
      Defaults to 'fullpath'.
    - initial_P_guess_kbar (float, optional): Initial guess for internal pressure (Pinternal_MPa) in MPa. Defaults to 0.
    - tolerance (float, optional): Tolerance for pressure calculations. Defaults to 0.001.
    - method (str, optional): The numerical integration method to use for change in FI radius. Can be 'RK1' (also 'Euler'), 'RK2' (also 'Heun'), 'RK3', or 'RK4'.
      Defaults to 'RK4'.
    - update_b (bool, optional): Whether to update 'b' during the ascent. Defaults to False.

    Returns:
    - pandas.DataFrame: A DataFrame containing the simulation results, including time, depth, pressure, radius changes,
      and CO2 density.

    """

    D, Pexternal_steps, dt_s = calculate_DPdt(ascent_rate_ms=ascent_rate_ms, crustal_model_config=crustal_model_config,
                                            D_initial_km=depth_path_ini_fin_step[0], D_final_km=depth_path_ini_fin_step[1],
                                            D_step=depth_path_ini_fin_step[2],
                                            initial_P_guess_kbar=initial_P_guess_kbar, tolerance=tolerance)
    Pinternal_MPa = Pexternal_steps[0]

    CO2_dens_initial, CO2_mass_initial = get_initial_CO2(R_m=R_m, T_K=T_K, P_MPa=Pinternal_MPa, EOS=EOS)

    
    results = pd.DataFrame([{'Time(s)': 0,
                            'Step':0,
                            'dt(s)':0,
                        'Pexternal(MPa)': Pinternal_MPa,
                        'Pinternal(MPa)': Pinternal_MPa,
                        'dR/dt(m/s)': calculate_dR_dt(R_m=R_m, b_m=b_m, Pinternal_MPa=Pinternal_MPa, Pexternal_MPa=Pinternal_MPa, T_K=T_K),
                        'Fi_radius(μm)': R_m*10**6,
                        'b (distance to xtal rim -μm)':b_m*10**6,
                        '\u0394R/R0 (fractional change in radius)':np.nan,
                        'CO2_dens_gcm3': CO2_dens_initial,
                        'Depth(km)':D.iloc[0]}], index=range(len(Pexternal_steps)))

   
    for i in range(1,len(Pexternal_steps)):
        
        Pexternal = Pexternal_steps[i]

        R_m,dR_dt = get_R(R_m=R_m,b_m=b_m,T_K=T_K,Pinternal_MPa=Pinternal_MPa,Pexternal_MPa=Pexternal,dt_s=dt_s,method=method)
        CO2_dens_new,P_new = get_CO2dens_P(R_m=R_m,T_K=T_K,CO2_mass=CO2_mass_initial,EOS=EOS)

        Pinternal_MPa = P_new

        if update_b==True:
            b_m=1000*R_m
        
        results.loc[i] = [dt_s*i, i, dt_s, Pexternal, Pinternal_MPa, dR_dt, R_m * 10 ** 6, b_m * 10 ** 6,
                    (R_m * 10 ** 6 - results.loc[0, 'Fi_radius(μm)']) / results.loc[0, 'Fi_radius(μm)'],
                    CO2_dens_new, D.iloc[i]]

    if report_results == 'startendonly':
        results.drop(index=list(range(1, results.shape[0] - 1)), inplace=True)  # Drop all rows except first and last

    if report_results == 'endonly':
        results.drop(index=list(range(0, results.shape[0] - 1)), inplace=True)  # Drop all rows except last

    if plotfig==True:
        fig, (ax0,ax1) = plt.subplots(1,2, figsize=(10,3))
        ax0.plot(results['Depth(km)'],results['\u0394R/R0 (fractional change in radius)'],marker='s',label=f"Ascent Rate = {ascent_rate_ms} m/s")
        ax0.set_xlim([depth_path_ini_fin_step[0],depth_path_ini_fin_step[1]])
        ax0.set_xlabel("Depth (km)")
        ax0.set_ylabel('\u0394R/R0 (fractional change in radius)')

        ax1.plot(results['Depth(km)'],results['CO2_dens_gcm3'],marker='s',label=f"Ascent Rate = {ascent_rate_ms} m/s")
        ax1.set_xlim([depth_path_ini_fin_step[0],depth_path_ini_fin_step[1]])
        ax1.set_xlabel("Depth (km)")
        ax1.set_ylabel("CO$_2$ density (g/cm$^{3}$)")
        ax0.legend(loc='best')
        ax1.legend(loc='best')
        fig.tight_layout()
        plt.show()

    return results

# This function is to model stretching at fixed External Pressure (e.g., during stalling or upon eruption)
def stretch_at_constant_Pext(*,R_m,b_m,T_K,EOS='SW96',Pinternal_MPa,Pexternal_MPa,totaltime_s,steps,method='RK4',report_results='fullpath',plotfig=False,update_b=False):
    """
    Simulate the stretching of a CO2 fluid inclusion (FI) under constant external pressure (e.g., quenching or storage).

    Parameters:
    - R_m (float): The initial radius of the fluid inclusion (FI), in meters.
    - b_m (float): The initial distance to a crystal defect (rim, crack, etc) from the FI center, in meters.
    - T_K (float): The temperature, in Kelvin.
    - Pinternal_MPa (float): The initial internal pressure of the FI, in MegaPascals (MPa).
    - Pexternal_MPa (float): The constant external pressure applied to the FI, in MegaPascals (MPa).
    - totaltime_s (float): The total simulation time, in seconds.
    - steps (int): The number of simulation steps.
    - EOS (str): The equation of state (EOS) to use for density calculations. Can be one of: 'ideal' (ideal gas),
      'SW96' (Span and Wagner 1996), or 'SP94' (Sterner and Pitzer 1994).
    - method (str, optional): The numerical integration method to use for change in FI radius. Can be 'RK1' (also 'Euler'), 'RK2' (also 'Heun'), 'RK3', or 'RK4'.
      Defaults to 'RK4'.
    - report_results (str, optional): The type of results to report. Can be 'fullpath' (all steps reported), 'startendonly' (only initial and end steps), or 'endonly' (only last step).
      Defaults to 'fullpath'.
    - plotfig (bool, optional): Whether to plot figures showing the changes in time, radius, and CO2 density. Defaults to False.
    - update_b (bool, optional): Whether to update 'b' during the simulation. Defaults to False.

    Returns:
    - pandas.DataFrame: A DataFrame containing the simulation results, including time, pressure, radius changes, and CO2 density.

    Raises:
    - ValueError: If an unsupported EOS is specified.
    """

    CO2_dens_initial,CO2_mass_initial=get_initial_CO2(R_m=R_m,T_K=T_K,P_MPa=Pinternal_MPa,EOS=EOS)

    results = pd.DataFrame([{'Time(s)': 0,
                             'Step':0,
                             'dt(s)':0,
                            'Pexternal(MPa)': Pexternal_MPa,
                            'Pinternal(MPa)': Pinternal_MPa,
                            'dR/dt(m/s)': calculate_dR_dt(R_m=R_m, b_m=b_m, Pinternal_MPa=Pinternal_MPa, Pexternal_MPa=Pexternal_MPa, T_K=T_K),
                            'Fi_radius(μm)': R_m*10**6,
                            'b (distance to xtal rim -μm)':b_m*10**6,
                            '\u0394R/R0 (fractional change in radius)':0,
                            'CO2_dens_gcm3': CO2_dens_initial}], index=range(steps))

    dt_s=totaltime_s/steps
    
    for step in range(1,steps):

        R_new,dR_dt = get_R(R_m=R_m,b_m=b_m,T_K=T_K,Pinternal_MPa=Pinternal_MPa,Pexternal_MPa=Pexternal_MPa,dt_s=dt_s,method=method)

        CO2_dens_new,P_new = get_CO2dens_P(R_m=R_new,T_K=T_K,CO2_mass=CO2_mass_initial,EOS=EOS)

        Pinternal_MPa = P_new
        R_m=R_new

        if update_b==True:
            b_m=1000*R_m
        
        results.loc[step] = [step * dt_s, step, dt_s, Pexternal_MPa, Pinternal_MPa, dR_dt, R_m * 10 ** 6, b_m * 10 ** 6,
                            (R_m * 10 ** 6 - results.loc[0, 'Fi_radius(μm)']) / results.loc[0, 'Fi_radius(μm)'],
                                CO2_dens_new]

    if report_results == 'startendonly':
        results.drop(index=list(range(1, results.shape[0] - 1)), inplace=True)  # Drop all rows except first and last

    if report_results == 'endonly':
        results.drop(index=list(range(0, results.shape[0] - 1)), inplace=True)  # Drop all rows except last

    if plotfig==True:
        if totaltime_s < 60:
            x_time = results['Time(s)']
            xlabel = 'Time(s)'
        elif 60 <= totaltime_s < 3600:
            x_time = results['Time(s)'] / 60
            xlabel = 'Time(min)'
            results[xlabel]=x_time
        elif 3600 <= totaltime_s < 86400:
            x_time = results['Time(s)'] / 3600
            xlabel = 'Time(hr)'
            results[xlabel]=x_time
        elif 86400 <= totaltime_s < 31536000:
            x_time = results['Time(s)'] / (3600 * 24)
            xlabel = 'Time(days)'
            results[xlabel]=x_time
        elif totaltime_s >= 31536000:
            x_time = results['Time(s)'] / (3600 * 24 * 365)
            xlabel = 'Time(years)'
            results[xlabel]=x_time

        fig, (ax0,ax1) = plt.subplots(1,2, figsize=(10,3))
        ax0.plot(x_time,results['\u0394R/R0 (fractional change in radius)'],marker='s')
        ax0.set_xlabel(xlabel)
        ax0.set_ylabel("\u0394R/R0 (fractional change in radius)")

        ax1.plot(x_time,results['CO2_dens_gcm3'],marker='s')
        ax1.set_xlabel(xlabel)
        ax1.set_ylabel("CO2_density_gmL")
        fig.tight_layout()
        plt.show()

    return results

# This function can loop through different R and b value sets using stretch at constant Pext

def loop_R_b_constant_Pext(*,R_m_values, b_m_values, T_K, EOS, Pinternal_MPa, Pexternal_MPa, totaltime_s, steps, T4endcalc_PD, method='RK4',
                            plotfig=False, crustal_model_config=config_crustalmodel(crust_dens_kgm3=2750)):

    """
    Perform multiple simulations under constant external pressure with various R and b values.

    Parameters:
    - R_m_values (list): A list of initial radius values for fluid inclusions (FI), in meters.
    - b_m_values (list): A list of initial distance values to a crystal defect (rim, crack, etc) from the FI center, in meters.
    - T_K (float): The temperature, in Kelvin.
    - EOS (str): The equation of state (EOS) to use for density calculations. Can be one of: 'ideal' (ideal gas),
      'SW96' (Span and Wagner 1996), or 'SP94' (Sterner and Pitzer 1994).
    - Pinternal_MPa (float): The initial internal pressure of the FI, in MegaPascals (MPa).
    - Pexternal_MPa (float): The constant external pressure applied to the FI, in MegaPascals (MPa).
    - totaltime_s (float): The total simulation time, in seconds.
    - steps (int): The number of simulation steps.
    - T4endcalc_PD (float): The temperature at which to calculate the depths (Kelvin) at the end of the simulations.
    - method (str, optional): The numerical integration method to use for change in FI radius. Can be 'RK1' (also 'Euler'), 'RK2' (also 'Heun'), 'RK3', or 'RK4'.
      Defaults to 'RK4'.
    - plotfig (bool, optional): Whether to plot figures showing the changes in time, radius, and CO2 density. Defaults to False.
    - crustal_model_config (dict, optional): Configuration parameters for the crustal model. Defaults to a predefined 
        configuration with a crustal density of 2750 kg/m³.

    Returns:
    - dict: A dictionary containing simulation results for varying R and b values. The keys are in the format 'R{index}_b{index}',
          where 'index' corresponds to the index of R_values and b_values, respectively. The values are DataFrames with 
          simulation results.
    """

    results_dict = {}
    for idx_R, R in enumerate(R_m_values):  # Use enumerate to get the index of R_values
        R_key = f'R{idx_R}'  # Use 'R' followed by the index
        results_dict[R_key] = {}

        for idx_b, b in enumerate(b_m_values):  # Use enumerate to get the index of b_values
            b_key = f'b{idx_b}'  # Use 'b' followed by the index
            results = stretch_at_constant_Pext(R_m=R, b_m=b, T_K=T_K, Pinternal_MPa=Pinternal_MPa, Pexternal_MPa=Pexternal_MPa,
                                              totaltime_s=totaltime_s, steps=steps, EOS=EOS, method=method,
                                              plotfig=plotfig)
            results['Calculated depths (km)_StorageT'] = convert_pressure_to_depth(
                P_kbar=results['Pinternal(MPa)'] / 100,
                crust_dens_kgm3=crustal_model_config.crust_dens_kgm3, g=9.81,
                d1=crustal_model_config.d1, d2=crustal_model_config.d2,
                rho1=crustal_model_config.rho1, rho2=crustal_model_config.rho2, rho3=crustal_model_config.rho3,
                model=crustal_model_config.model)
            results['Calculated P from rho (MPa)_TrappingT'] = calculate_P_for_rho_T(
                EOS='SW96', CO2_dens_gcm3=results['CO2_dens_gcm3'], T_K=T4endcalc_PD + 273.15)['P_MPa']

            results['Calculated depths (km)_TrappingT'] = convert_pressure_to_depth(
                P_kbar=results['Calculated P from rho (MPa)_TrappingT'] / 100,
                crust_dens_kgm3=crustal_model_config.crust_dens_kgm3, g=9.81,
                d1=crustal_model_config.d1, d2=crustal_model_config.d2,
                rho1=crustal_model_config.rho1, rho2=crustal_model_config.rho2, rho3=crustal_model_config.rho3,
                model=crustal_model_config.model)

            results_dict[R_key][b_key] = results

    return results_dict
